<?php

namespace Drupal\mcapi\Plugin\DevelGenerate;

use Drupal\mcapi\Functions;
use Drupal\mcapi\Entity\Transaction;
use Drupal\mcapi\Entity\Wallet;
use Drupal\devel_generate\DevelGenerateBase;
use Drupal\Core\Batch\BatchBuilder;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Form\FormStateInterface;
use Drupal\Core\Plugin\ContainerFactoryPluginInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Provides a DevelGenerate plugin.
 *
 * @DevelGenerate(
 *   id = "mc_transaction",
 *   label = @Translation("transactions"),
 *   description = @Translation("Generate a given number of transactions..."),
 *   url = "transaction",
 *   permission = "administer devel_generate",
 *   settings = {
 *     "num" = 100,
 *     "kill" = TRUE,
 *     "type" = "default"
 *   }
 * )
 */
class TransactionDevelGenerate extends DevelGenerateBase implements ContainerFactoryPluginInterface {

  const MAX = 100;

  /**
   * The transaction storage.
   *
   * @var \Drupal\Core\Entity\EntityStorageInterface
   */
  protected $transactionStorage;

  /**
   * All the wallet ids
   *
   * @var array
   */
  private $wids;

  /**
   * @param array $configuration
   * @param type $plugin_id
   * @param array $plugin_definition
   * @param EntityTypeManagerInterface $entity_type_manager
   * @throws \Exception
   */
  public function __construct(array $configuration, $plugin_id, array $plugin_definition, EntityTypeManagerInterface $entity_type_manager) {
    parent::__construct($configuration, $plugin_id, $plugin_definition);
    $this->transactionStorage = $entity_type_manager->getStorage('mc_transaction');
    $this->wids = $entity_type_manager->getstorage('mc_wallet')->getQuery()->accessCheck(FALSE)->execute();

    if (count($this->wids) < 2) {
      throw new \Exception('Not enough wallets to make a transaction.');
    }
    $this->requestTime = \Drupal::time()->getRequestTime();
  }

  /**
   * {@inheritDoc}
   */
  public static function create(ContainerInterface $container, array $configuration, $plugin_id, $plugin_definition) {
    return new static(
      $configuration, $plugin_id, $plugin_definition,
      $container->get('entity_type.manager')
    );
  }

  /**
   * {@inheritDoc}
   */
  public function settingsForm(array $form, FormStateInterface $form_state) {
    $form['kill'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Delete existing transactions of the given type.'),
      '#default_value' => $this->getSetting('kill'),
    ];
    $form['num'] = [
      '#type' => 'number',
      '#title' => $this->t('How many transactions would you like to generate?'),
      '#default_value' => $this->getSetting('num'),
      '#required' => TRUE,
      '#min' => 0,
    ];
    // @todo add 'all' types option.
    $form['type'] = [
      '#title' => $this->t('What type of transactions'),
      '#type' => 'select',
      '#options' => Functions::entityOptions('mc_type'),
      '#default_value' => $this->getSetting('type'),
      '#required' => TRUE,
      '#min' => 0,
    ];
    $form['#redirect'] = FALSE;
    return $form;
  }

  /**
   * {@inheritDoc}
   */
  protected function generateElements(array $values) {
    $this->settings = $values + $this->settings;
    if ($this->getSetting('num') <= static::MAX) {
      if ($this->getSetting('kill')) {
        $this->contentKill();
      }

      for ($i = 0; $i < $this->getSetting('num'); $i++) {
        $this->develGenerateTransactionAdd();
      }
      if (function_exists('drush_log') && $i % drush_get_option('feedback', 1000) == 0) {
        drush_log(dt('Completed @feedback transactions ', ['@feedback' => drush_get_option('feedback', 1000)], 'ok'));
      }
      $this->setMessage($this->t('Created @num transactions.', ['@num' => $this->getSetting('num')]));
    }
    else {
      // setFile doesn't include the file yet and devel_generate_batch_finished must be callable
      \Drupal::moduleHandler()->loadInclude('devel_generate', 'batch.inc');
      $batch_builder = (new BatchBuilder())
       ->setTitle($this->t('Generating transactions'))
       ->setFile(\Drupal::service('extension.path.resolver')->getPath('module', 'devel_generate') . '/devel_generate.batch.inc')
       ->setFinishCallback('devel_generate_batch_finished');
      // Add the kill operation.
      if ($this->getSetting('kill')) {
        $batch_builder->addOperation('devel_generate_operation', [$this, 'batchContentKill', []]);
      }

      $batches = ceil($this->getSetting('num')/static::MAX);
      for ($num = 0; $num < $batches; $num++) {
        $batch_builder->addOperation('devel_generate_operation', [$this, 'batchContentAddTransactions', []]);
      }
      batch_set($batch_builder->toArray());
    }
  }

  /**
   * batch callback to create one transaction.
   */
  public function batchContentAddTransactions($values, &$context) {
    $i = 0;
    if (!isset($context['results']['num'])) {
      $context['results']['num'] = 0;
    }
    while ($context['results']['num'] < $this->getSetting('num') and $i < Self::MAX) {
      $this->develGenerateTransactionAdd();
      $context['results']['num']++;
      $i++;
    }
  }

  /**
   * batch callback to create one transaction.
   */
  public function batchContentKill($values, &$context) {
    $this->contentKill();
  }

  /**
   * Deletes all transactions .
   *
   * @param string $type
   *   The type of transactions to delete
   *
   * @note Loads all transactions into memory at the same time.
   */
  public function contentKill() {
    $xids = $this->transactionStorage->getQuery()->accessCheck(FALSE)
      ->condition('type', $this->getSetting('type'))
      ->execute();
    if (!empty($xids)) {
      $transactions = Transaction::loadMultiple($xids);
      $this->transactionStorage->delete($transactions);
      $this->setMessage($this->t('Deleted %count transactions.', array('%count' => count($xids))));
    }
  }

  /**
   * Create one transaction. Used by both batch and non-batch code branches.
   *
   * @note this may attempt to send a email for pending transactions.
   */
  public function develGenerateTransactionAdd() {
    list($w1, $w2) = $this->get2RandWalletIds();
    if (!$w2) {
      return;
    }
    $creator = Wallet::load($w1)->getOwner();
    $props = [
      'payer' => $w1,
      'payee' => $w2,
      // Transactions of type 'inherit' don't show in the default view.
      'type' => $this->getSetting('type') ?: '3rdparty',
      'creator' => $creator->id(),
      'description' => $this->getRandom()->sentences(1),
      'quant' => rand(10, 250)
    ];
    $transaction = Transaction::create($props);
    // Populate additional fields
    $this->populateFields($transaction);

    // We don't want all transactions created at the same time!
    // Make created time between the newest involved wallet creation and now.
    $latest = 0;
    foreach ($transaction->allWallets() as $wallet) {
      $latest = max($latest, $wallet->created->value);
    }
    $transaction->created->value = rand($latest, $this->requestTime);

    try {
      // Validation is needed not least because that is when the transaction is assembled.
      $transaction->validate();
    }
    catch (\Exception $e) {}
    $transaction->state->value = key($transaction->type->entity->states);
    \Drupal::service('account_switcher')->switchTo($creator);
    $transaction->save();
    \Drupal::service('account_switcher')->switchBack();

    if ($transaction->state->value == 'pending') {
      $transaction->original = clone($transaction);// needed to save it again
      // Signatures already exist because they were created in presave phase.
      foreach ($transaction->signatures as $uid => $signed) {
        // Leave 1 in 5 signatures unsigned.
        if (rand(0, 5) > 0) {
          $transaction->signatures->sign($uid);
          $transaction->save();
        }
      }
      // NB this could generate pending emails.
      if (isset($resave)) {
        $transaction->save();
      }
    }
  }

  /**
   * Get two random wallets
   *
   * @return int[]
   *   2 wallet ids
   */
  public function get2RandWalletIds() {
    $wids = $this->wids;
    shuffle($wids);
    return array_slice($wids, -2);
  }

  /**
   * {@inheritDoc}
   */
  public function validateDrushParams($args, $options = []) {
    $values['kill'] = drush_get_option('kill');
    $values['type'] = drush_get_option('type');
    $values['num'] = array_shift($args);
    return $values;
  }

}
